@@ -1,6 +1,14 @@
+import os
+import pickle
+
 import requests
 from bs4 import BeautifulSoup
 
+from .requester import requester
+
+
+_FANDOMS = None
+_LANGUAGES = None
 
 class LoginError(Exception):
     def __init__(self, message, errors=[]):
@@ -47,7 +55,6 @@ class HTTPError(Exception):
         super().__init__(message)
         self.errors = errors
 
-
 class Query:
     def __init__(self):
         self.fields = []
@@ -94,8 +101,89 @@ class Constraint:
 
     def __str__(self):
         return self.string
+    
+def set_rqtw(value):
+    """Sets the requests per time window parameter for the AO3 requester"""
+    requester.setRQTW(value)
+    
+def set_timew(value):
+    """Sets the time window parameter for the AO3 requester"""
+    requester.setTimeW(value)
         
+def limit_requests(limit=True):
+    """Toggles request limiting"""
+    #! NOT TESTED WITH THREADING
+    
+    if limit:
+        #! AO3's code throttles requests at 60rpm, but that doesn't seem to work.
+        requester.setRQTW(110)
+    else:
+        requester.setRQTW(-1)
+    
+def load_fandoms():
+    """Loads fandoms into memory
 
+    Raises:
+        FileNotFoundError: No resource was found
+    """
+    
+    global _FANDOMS
+    
+    fandom_path = os.path.join(os.path.dirname(__file__), "resources", "fandoms")
+    if not os.path.isdir(fandom_path):
+        raise FileNotFoundError("No fandom resources have been downloaded. Try AO3.extra.download()")
+    files = os.listdir(fandom_path)
+    _FANDOMS = []
+    for file in files:
+        with open(os.path.join(fandom_path, file), "rb") as f:
+            _FANDOMS += pickle.load(f)
+            
+def load_languages():
+    """Loads languages into memory
+
+    Raises:
+        FileNotFoundError: No resource was found
+    """
+    
+    global _LANGUAGES
+    
+    language_path = os.path.join(os.path.dirname(__file__), "resources", "languages")
+    if not os.path.isdir(language_path):
+        raise FileNotFoundError("No language resources have been downloaded. Try AO3.extra.download()")
+    files = os.listdir(language_path)
+    _LANGUAGES = []
+    for file in files:
+        with open(os.path.join(language_path, file), "rb") as f:
+            _LANGUAGES += pickle.load(f)
+            
+def get_languages():
+    """Returns all available languages"""
+    return _LANGUAGES[:]
+
+def search_fandom(fandom_string):
+    """Searches for a fandom that matches the given string
+
+    Args:
+        fandom_string (str): query string
+
+    Raises:
+        UnloadedError: load_fandoms() wasn't called
+        UnloadedError: No resources were downloaded
+
+    Returns:
+        list: All results matching 'fandom_string'
+    """
+    
+    if _FANDOMS is None:
+        raise UnloadedError("Did you forget to call AO3.utils.load_fandoms()?")
+    if _FANDOMS == []:
+        raise UnloadedError("Did you forget to download the required resources with AO3.extra.download()?")
+    results = []
+    for fandom in _FANDOMS:
+        if fandom_string.lower() in fandom.lower():
+            results.append(fandom)
+    return results
+        
 def workid_from_url(url):
     """Get the workid from an archiveofourown.org website url
 
@@ -158,10 +246,12 @@ def comment(chapterid, comment_text, sess, oneshot=False, commentid=None, email=
         else:
             referer = f"https://archiveofourown.org/chapters/{chapterid}"
             
-        soup = sess.request(referer)            
+        soup = sess.request(referer)   
         pseud = soup.find("input", {"name": "comment[pseud_id]"})
         if pseud is None:
             pseud = soup.find("select", {"name": "comment[pseud_id]"})
+            if pseud is None:
+                raise PseudError("Couldn't find your pseud's id")
             pseud_id = None
             for option in pseud.findAll("option"):
                 if "selected" in option.attrs and option.attrs["selected"] == "selected":
@@ -246,7 +336,6 @@ def delete_comment(commentid, session):
                 raise PermissionError("You don't have permission to do this")
     raise UnexpectedResponseError("An unexpected error has occurred")
             
-    
 def kudos(workid, session):
     """Leave a 'kudos' in a specific work
 
@@ -290,3 +379,44 @@ def kudos(workid, session):
         raise UnexpectedResponseError(f"Unexpected json received:\n"+str(json))
     else:
         raise UnexpectedResponseError(f"Unexpected HTTP status code received ({response.status_code})")
+    
+def subscribe(workid, worktype, session, unsubscribe=False, subid=None):
+    """Subscribes to a work. Be careful, you can subscribe to a work multiple times
+
+    Args:
+        workid (int/str): ID of the work
+        worktype (str): Type of the work (Series/Work/User)
+        session (AO3.Session): Session object
+        unsubscribe (bool, optional): Unsubscribe instead of subscribing. Defaults to False.
+        subid (str/int, optional): Subscription ID, used when unsubscribing. Defaults to None.
+
+    Raises:
+        AuthError: Invalid auth token
+        AuthError: Invalid session
+        InvalidIdError: Invalid workid / worktype
+        InvalidIdError: Invalid subid
+    """
+    
+    if not session.is_authed:
+        raise AuthError("Invalid session")
+    
+    data = {
+        "authenticity_token": session.authenticity_token,
+        "subscription[subscribable_id]": workid,
+        "subscription[subscribable_type]": worktype.capitalize()
+    }
+
+    url = f"https://archiveofourown.org/users/{session.username}/subscriptions"
+    if unsubscribe:
+        if subid is None:
+            raise InvalidIdError("When unsubscribing, subid cannot be None")
+        url += f"/{subid}"
+        data["_method"] = "delete"
+    req = session.session.post(url, data=data, allow_redirects=False)
+    if unsubscribe:
+        return req
+    if req.status_code == 302:
+        if req.headers["Location"] == "https://archiveofourown.org/auth_error":
+            raise AuthError("Invalid authentication token. Try calling session.refresh_auth_token()")
+    else:
+        raise InvalidIdError(f"Invalid workid / worktype")
