@@ -1,11 +1,11 @@
-from datetime import date
+from datetime import datetime
 from functools import cached_property
 
-import requests
 from bs4 import BeautifulSoup
 
-from . import utils
+from . import threadable, utils
 from .comments import Comment
+from .requester import requester
 
 
 class Work:
@@ -13,12 +13,13 @@ class Work:
     AO3 work object
     """
 
-    def __init__(self, workid, session=None):
+    def __init__(self, workid, session=None, load=True):
         """Creates a new AO3 work object
 
         Args:
             workid (int): AO3 work ID
             session (AO3.Session, optional): Used to access restricted works
+            load (bool, optional): If true, the work is loaded on initialization. Defaults to True.
 
         Raises:
             utils.InvalidIdError: Raised if the work wasn't found
@@ -28,13 +29,66 @@ class Work:
         self.chapter_ids = []
         self.chapter_names = []
         self.workid = workid
-        self.soup = self.request("https://archiveofourown.org/works/%i?view_adult=true"%workid)
-        if "Error 404" in self.soup.text:
+        self._soup = None
+        if load:
+            self.reload()
+            
+    def __repr__(self):
+        try:
+            return f"<Work [{self.title}]>"
+        except:
+            return f"<Work [{self.workid}]>"
+    
+    def __eq__(self, other):
+        return isinstance(other, Work) and other.workid == self.workid
+    
+    def __getstate__(self):
+        d = {}
+        for attr in self.__dict__:
+            if isinstance(self.__dict__[attr], BeautifulSoup):
+                d[attr] = (self.__dict__[attr].encode(), True)
+            else:
+                d[attr] = (self.__dict__[attr], False)
+        return d
+                
+    def __setstate__(self, d):
+        for attr in d:
+            value, issoup = d[attr]
+            if issoup:
+                self.__dict__[attr] = BeautifulSoup(value, "lxml")
+            else:
+                self.__dict__[attr] = value
+        
+    @threadable.threadable
+    def reload(self):
+        """
+        Loads information about this work.
+        This function is threadable.
+        """
+        
+        for attr in self.__class__.__dict__:
+            if isinstance(getattr(self.__class__, attr), cached_property):
+                if attr in self.__dict__:
+                    delattr(self, attr)
+        
+        self._soup = self.request("https://archiveofourown.org/works/%i?view_adult=true"%self.workid)
+        if "Error 404" in self._soup.text:
             raise utils.InvalidIdError("Cannot find work")
+        self.load_chapters()
+        
+    def set_session(self, session):
+        """Sets the session used to make requests for this work
+
+        Args:
+            session (AO3.Session/AO3.GuestSession): session object
+        """
+        
+        self._session = session 
 
     def get_chapter_text(self, chapter):
         """Gets the chapter text from the specified chapter.
-        Work.load_chapters() must be called first.
+        If chapter=-1 then it gets the text from all chapters.
+        The work must be loaded first.
 
         Args:
             chapter (int): Work chapter
@@ -46,25 +100,79 @@ class Work:
             str: Chapter text
         """
         
+        if not self.loaded:
+            raise utils.UnloadedError("Work isn't loaded. Have you tried calling Work.reload()?")
+        
+        text = ""
+        if chapter > 0 and chapter <= self.chapters and not self.oneshot:
+            chapter_html = self.request("https://archiveofourown.org/works/%i/chapters/%s?view_adult=true"%(self.workid, self.chapter_ids[chapter-1]))
+            div = chapter_html.find("div", {"role": "article"})
+            for p in div.findAll("p"):
+                text += p.getText().replace("\n", "") + "\n"
+        elif self.oneshot:
+            div = self._soup.find("div", {"role": "article"})
+            for p in div.findAll("p"):
+                text += p.getText().replace("\n", "") + "\n"
+        else:
+            work_html = self.request("https://archiveofourown.org/works/%i/?view_adult=true&view_full_work=true"%(self.workid))
+            chapters_div = work_html.find("div", {"id": "chapters"})
+            for div in chapters_div.findAll("div", {"role": "article"}):
+                for p in div.findAll("p"):
+                    text += p.getText().replace("\n", "") + "\n"
+        return text
+    
+    def get_chapter_images(self, chapter):
+        """Gets the chapter text from the specified chapter.
+        Work.load_chapters() must be called first.
+
+        Args:
+            chapter (int): Work chapter
+
+        Raises:
+            utils.UnloadedError: Raises this error if the chapters aren't loaded
+
+        Returns:
+            tuple: Pairs of image urls and the paragraph number
+        """
+        
+        if not self.loaded:
+            raise utils.UnloadedError("Work isn't loaded. Have you tried calling Work.reload()?")
         if chapter > 0 and chapter <= self.chapters and self.chapters > 1:
             if len(self.chapter_ids) == self.chapters:
                 chapter_html = self.request("https://archiveofourown.org/works/%i/chapters/%s?view_adult=true"%(self.workid, self.chapter_ids[chapter-1]))
                 div = chapter_html.find("div", {'role': 'article'})
-                return str(BeautifulSoup.getText(div))
+                images = []
+                line = 0
+                for p in div.findAll("p"):
+                    line += 1
+                    for img in p.findAll("img"):
+                        images.append((img.attrs["src"], line))
+                return tuple(images)
             else:
                 raise utils.UnloadedError("Work.load_chapters() must be called first")
 
         elif chapter == 1:
-            div = self.soup.find("div", {'role': 'article'})
-            return str(BeautifulSoup.getText(div))
+            div = self._soup.find("div", {'role': 'article'})
+            images = []
+            line = 0
+            for p in div.findAll("p"):
+                line += 1
+                for img in p.findAll("img"):
+                    images.append((img.attrs["src"], line))
+            return tuple(images)
         else:
             raise utils.UnloadedError("Work.load_chapters() must be called first")
     
-    def load_chapters(self, session=None):
-        """
-        Loads the urls for all chapters
+    def load_chapters(self):
+        """Loads the urls for all chapters
+
+        Raises:
+            utils.UnloadedError: Work isn't loaded
+            utils.AuthError: This is a private work
         """
         
+        if not self.loaded:
+            raise utils.UnloadedError("Work isn't loaded. Have you tried calling Work.reload()?")
         if not self.oneshot:
             navigate = self.request("https://archiveofourown.org/works/%i/navigate?view_adult=true"%self.workid)
             all_chapters = navigate.find("ol", {'class': 'chapter index group'})
@@ -73,8 +181,8 @@ class Work:
             self.chapter_ids = []
             self.chapter_names = []
             for chapter in all_chapters.findAll("li"):
-                self.chapter_ids.append(chapter.a['href'].split("/")[-1])
-                self.chapter_names.append(chapter.a.string)
+                self.chapter_ids.append(str(chapter.a['href'].split("/")[-1]))
+                self.chapter_names.append(str(chapter.a.string))
         else:
             self.chapter_ids = [""]
             self.chapter_names = [self.title]
@@ -93,11 +201,13 @@ class Work:
             bytes: File content
         """
         
-        download_btn = self.soup.find("li", {"class": "download"})
+        if not self.loaded:
+            raise utils.UnloadedError("Work isn't loaded. Have you tried calling Work.reload()?")
+        download_btn = self._soup.find("li", {"class": "download"})
         for download_type in download_btn.findAll("li"):
-            if download_type.a.getText() == filetype:
+            if download_type.a.getText() == filetype.upper():
                 url = f"https://archiveofourown.org/{download_type.a.attrs['href']}"
-                req = requests.get(url)
+                req = self.get(url)
                 if req.status_code == 429:
                     raise utils.HTTPError("We are being rate-limited. Try again in a while or reduce the number of requests")
                 if not req.ok:
@@ -105,6 +215,45 @@ class Work:
                 return req.content
         raise utils.UnexpectedResponseError(f"Filetype '{filetype}' is not available for download")
     
+    @threadable.threadable
+    def download_to_file(self, filename, filetype="PDF"):
+        """Downloads this work and saves it in the specified file.
+        This function is threadable.
+
+        Args:
+            filename (str): Name of the resulting file
+            filetype (str, optional): Desired filetype. Defaults to "PDF".
+
+        Raises:
+            utils.DownloadError: Raised if there was an error with the download
+            utils.UnexpectedResponseError: Raised if the filetype is not available for download
+        """
+        with open(filename, "wb") as file:
+            file.write(self.download(filetype))
+            
+    @property
+    def metadata(self):
+        metadata = {
+            "authors": list(map(lambda author: author.username, self.authors)),
+            "bookmarks": self.bookmarks,
+            "categories": self.categories,
+            "characters": self.characters,
+            "date_published": str(self.date_published),
+            "date_updated": str(self.date_updated),
+            "fandoms": self.fandoms,
+            "hits": self.hits,
+            "kudos": self.kudos,
+            "language": self.language,
+            "rating": self.rating,
+            "relationships": self.relationships,
+            "series": list(map(lambda series: series.name, self.series)),
+            "summary": self.summary,
+            "tags": self.tags,
+            "title": self.title,
+            "warnings": self.warnings
+        }
+        return metadata
+    
     def get_comments(self, chapter=None, maximum=None):
         """Returns a list of all threads of comments in the specified chapter. This operation can take a very long time.
         Because of that, it is recomended that you set a maximum number of comments. 
@@ -123,6 +272,8 @@ class Work:
             list: List of comments
         """
         
+        if not self.loaded:
+            raise utils.UnloadedError("Work isn't loaded. Have you tried calling Work.reload()?")
         if self.oneshot:
             chapter_id = self.workid
         else:
@@ -135,7 +286,8 @@ class Work:
             
             chapter_id = self.chapter_ids[chapter-1]
             
-        url = f"https://archiveofourown.org/comments/show_comments?page=%d&chapter_id={chapter_id}"
+        string = "works" if self.oneshot else "chapters" 
+        url = f"https://archiveofourown.org/{string}/{chapter_id}?page=%d&show_comments=true&view_adult=true"
         soup = self.request(url%1)
         
         pages = 0
@@ -157,42 +309,103 @@ class Work:
                 if maximum is not None and len(comments) >= maximum:
                     return comments
                 id_ = int(li.attrs["id"][8:])
-                comments.append(Comment(id_, chapter_id))
+                comments.append(Comment(id_, chapter_id, oneshot=self.oneshot))
         return comments
     
-    def leave_kudos(self, session):
-        """Leave a 'kudos' in this work
+    @threadable.threadable
+    def subscribe(self):
+        """Subscribes to this work.
+        This function is threadable.
 
-        Args:
-            session (AO3.Session/AO3.GuestSession): session object
+        Raises:
+            utils.AuthError: Invalid session
+        """
+        
+        if self._session is None or not self._session.is_authed:
+            raise utils.AuthError("You can only subscribe to a work using an authenticated session")
+        
+        utils.subscribe(self.workid, "Work", self._session)
+        
+    @threadable.threadable
+    def unsubscribe(self):
+        """Unubscribes from this user.
+        This function is threadable.
+
+        Raises:
+            utils.AuthError: Invalid session
+        """
+        
+        if not self.is_subscribed:
+            raise Exception("You are not subscribed to this work")
+        if self._session is None or not self._session.is_authed:
+            raise utils.AuthError("You can only unsubscribe from a work using an authenticated session")
+        
+        utils.subscribe(self.workid, "Work", self._session, True, self._sub_id)
+        
+    @cached_property
+    def is_subscribed(self):
+        """True if you're subscribed to this work"""
+        
+        if self._session is None or not self._session.is_authed:
+            raise utils.AuthError("You can only get a user ID using an authenticated session")
+        
+        ul = self._soup.find("ul", {"class": "work navigation actions"})
+        input_ = ul.find("li", {"class": "subscribe"}).find("input", {"name": "commit", "value": "Unsubscribe"})
+        return input_ is not None
+    
+    @cached_property
+    def _sub_id(self):
+        """Returns the subscription ID. Used for unsubscribing"""
+        
+        if self._session is None or not self._session.is_authed:
+            raise utils.AuthError("You can only get a user ID using an authenticated session")
+        
+        ul = self._soup.find("ul", {"class": "work navigation actions"})
+        id_ = ul.find("li", {"class": "subscribe"}).form.attrs["action"].split("/")[-1]
+        return int(id_)
+    
+    @threadable.threadable
+    def leave_kudos(self):
+        """Leave a 'kudos' in this work.
+        This function is threadable.
 
         Raises:
             utils.UnexpectedResponseError: Unexpected response received
             utils.InvalidIdError: Invalid workid (work doesn't exist)
-            utils.AuthError: Invalid authenticity token
+            utils.AuthError: Invalid session or authenticity token
 
         Returns:
             bool: True if successful, False if you already left kudos there
         """
         
-        return utils.kudos(self.workid, session)
+        if self._session is None:
+            raise utils.AuthError("Invalid session")
+        return utils.kudos(self.workid, self._session)
     
-    def comment(self, chapter, comment_text, session, email="", name=""):
-        """Leaves a comment on this work
+    @threadable.threadable
+    def comment(self, chapter, comment_text, email="", name=""):
+        """Leaves a comment on this work.
+        This function is threadable.
 
         Args:
             chapter (int): Chapter number
             comment_text (str): Comment text
-            session (AO3.Session, optional): Session object. Defaults to None (posts anonimously).
 
         Raises:
             IndexError: Invalid chapter number
             utils.UnloadedError: Couldn't load chapter ids. Call Work.load_chapters() first
+            utils.AuthError: Invalid session
 
         Returns:
             requests.models.Response: Response object
         """
         
+        if not self.loaded:
+            raise utils.UnloadedError("Work isn't loaded. Have you tried calling Work.reload()?")
+        
+        if self._session is None:
+            raise utils.AuthError("Invalid session")
+        
         if chapter < 1 or chapter > self.chapters:
             raise IndexError(f"Invalid chapter number")
         
@@ -204,11 +417,35 @@ class Work:
         else:
             chapterid = self.chapter_ids[chapter-1]
             
-        return utils.comment(chapterid, comment_text, self.oneshot, email=email, name=name)
+        return utils.comment(chapterid, comment_text, self._session, self.oneshot, email=email, name=name)
+    
+    @property
+    def loaded(self):
+        """Returns True if this work has been loaded"""
+        return self._soup is not None
     
     @property
     def oneshot(self):
+        """Returns True if this work has only one chapter"""
         return self.chapters == 1
+    
+    @cached_property
+    def series(self):
+        """Returns the series this work belongs to"""
+        
+        from .series import Series
+        dd = self._soup.find("dd", {"class": "series"})
+        if dd is None:
+            return []
+        
+        s = []
+        for span in dd.find_all("span", {"class": "position"}):
+            seriesid = int(span.a.attrs["href"].split("/")[-1])
+            seriesname = span.a.getText()
+            series = Series(seriesid, self._session, False)
+            setattr(series, "name", seriesname)
+            s.append(series)
+        return s
 
     @cached_property
     def authors(self):
@@ -218,12 +455,15 @@ class Work:
             list: list of authors
         """
 
-        authors = self.soup.find_all("a", {'rel': 'author'})
+        from .users import User
+        authors = self._soup.find_all("h3", {'class': 'byline heading'})
+        formatted_authors = authors[0].text.replace("\n", "").split(", ")
         author_list = []
         if authors is not None:
-            for author in authors:
-                author_list.append(author.string.strip())
-            
+            for author in formatted_authors:
+                user = User(author, load=False)
+                author_list.append(user)
+
         return author_list
 
     @cached_property
@@ -234,7 +474,7 @@ class Work:
             int: number of chapters
         """
         
-        chapters = self.soup.find("dd", {'class': 'chapters'})
+        chapters = self._soup.find("dd", {'class': 'chapters'})
         if chapters is not None:
             return int(self.str_format(chapters.string.split("/")[0]))
         return 0
@@ -247,7 +487,7 @@ class Work:
             int: number of hits
         """
 
-        hits = self.soup.find("dd", {'class': 'hits'})
+        hits = self._soup.find("dd", {'class': 'hits'})
         if hits is not None:
             return int(self.str_format(hits.string))
         return 0
@@ -260,7 +500,7 @@ class Work:
             int: number of kudos
         """
 
-        kudos = self.soup.find("dd", {'class': 'kudos'})
+        kudos = self._soup.find("dd", {'class': 'kudos'})
         if kudos is not None:
             return int(self.str_format(kudos.string))
         return 0
@@ -273,7 +513,7 @@ class Work:
             int: number of comments
         """
 
-        comments = self.soup.find("dd", {'class': 'comments'})
+        comments = self._soup.find("dd", {'class': 'comments'})
         if comments is not None:
             return int(self.str_format(comments.string))
         return 0
@@ -286,7 +526,7 @@ class Work:
             int: number of words
         """
 
-        words = self.soup.find("dd", {'class': 'words'})
+        words = self._soup.find("dd", {'class': 'words'})
         if words is not None:
             return int(self.str_format(words.string))
         return 0
@@ -299,7 +539,7 @@ class Work:
             str: Language
         """
 
-        language = self.soup.find("dd", {'class': 'language'})
+        language = self._soup.find("dd", {'class': 'language'})
         if language is not None:
             return language.string.strip()
         else:
@@ -313,7 +553,7 @@ class Work:
             int: number of bookmarks
         """
 
-        bookmarks = self.soup.find("dd", {'class': 'bookmarks'})
+        bookmarks = self._soup.find("dd", {'class': 'bookmarks'})
         if bookmarks is not None:
             return int(self.str_format(bookmarks.string))
         return 0
@@ -326,9 +566,9 @@ class Work:
             str: work title
         """
 
-        title = self.soup.find("div", {'class': 'preface group'})
+        title = self._soup.find("div", {'class': 'preface group'})
         if title is not None:
-            return str(title.h2.string.strip())
+            return str(title.h2.text.strip())
         return ""
     
     @cached_property
@@ -339,22 +579,22 @@ class Work:
             datetime.date: publish date
         """
 
-        dp = self.soup.find("dd", {'class': 'published'}).string
-        return date(*list(map(int, dp.split("-"))))
+        dp = self._soup.find("dd", {'class': 'published'}).string
+        return datetime(*list(map(int, dp.split("-"))))
 
     @cached_property
     def date_updated(self):
         """Returns the date this work was last updated
 
         Returns:
-            datetime.date: update date
+            datetime.datetime: update date
         """
 
-        if self.chapters > 1:
-            du = self.soup.find("dd", {'class': 'status'}).string
-            return date(*list(map(int, du.split("-"))))
-        else:
-            return self.date_published
+        download = self._soup.find("li", {"class": "download"})
+        if download is not None and download.ul is not None:
+            timestamp = int(download.ul.a["href"].split("=")[-1])
+            return datetime.fromtimestamp(timestamp)
+        return datetime(self.date_published)
     
     @cached_property
     def tags(self):
@@ -364,7 +604,7 @@ class Work:
             list: List of tags
         """
 
-        html = self.soup.find("dd", {'class': 'freeform tags'})
+        html = self._soup.find("dd", {'class': 'freeform tags'})
         tags = []
         if html is not None:
             for tag in html.find_all("li"):
@@ -379,7 +619,7 @@ class Work:
             list: List of characters
         """
 
-        html = self.soup.find("dd", {'class': 'character tags'})
+        html = self._soup.find("dd", {'class': 'character tags'})
         characters = []
         if html is not None:
             for character in html.find_all("li"):
@@ -394,7 +634,7 @@ class Work:
             list: List of relationships
         """
         
-        html = self.soup.find("dd", {'class': 'relationship tags'})
+        html = self._soup.find("dd", {'class': 'relationship tags'})
         relationships = []
         if html is not None:
             for relationship in html.find_all("li"):
@@ -409,7 +649,7 @@ class Work:
             list: List of fandoms
         """
 
-        html = self.soup.find("dd", {'class': 'fandom tags'})
+        html = self._soup.find("dd", {'class': 'fandom tags'})
         fandoms = []
         if html is not None:
             for fandom in html.find_all("li"):
@@ -424,7 +664,7 @@ class Work:
             list: List of categories
         """
 
-        html = self.soup.find("dd", {'class': 'category tags'})
+        html = self._soup.find("dd", {'class': 'category tags'})
         categories = []
         if html is not None:
             for category in html.find_all("li"):
@@ -439,7 +679,7 @@ class Work:
             list: List of warnings
         """
 
-        html = self.soup.find("dd", {'class': 'warning tags'})
+        html = self._soup.find("dd", {'class': 'warning tags'})
         warnings = []
         if html is not None:
             for warning in html.find_all("li"):
@@ -454,7 +694,7 @@ class Work:
             str: Rating
         """
 
-        html = self.soup.find("dd", {'class': 'rating tags'})
+        html = self._soup.find("dd", {'class': 'rating tags'})
         if html is not None:
             rating = html.a.string
             return rating
@@ -468,7 +708,7 @@ class Work:
             str: Summary
         """
 
-        div = self.soup.find("div", {'class': 'preface group'})
+        div = self._soup.find("div", {'class': 'preface group'})
         if div is None:
             return ""
         html = div.find("blockquote", {'class': 'userstuff'})
@@ -484,27 +724,31 @@ class Work:
             str: work URL
         """    
 
-        return "https://archiveofourown.org/works/%i"%self.workid      
+        return "https://archiveofourown.org/works/%i"%self.workid    
+    
+    def get(self, *args, **kwargs):
+        """Request a web page and return a Response object"""  
+        
+        if self._session is None:
+            req = requester.request("get", *args, **kwargs)
+        else:
+            req = requester.request("get", *args, **kwargs, session=self._session.session)
+        if req.status_code == 429:
+            raise utils.HTTPError("We are being rate-limited. Try again in a while or reduce the number of requests")
+        return req
 
     def request(self, url):
         """Request a web page and return a BeautifulSoup object.
 
         Args:
             url (str): Url to request
-            data (dict, optional): Optional data to send in the request. Defaults to {}.
 
         Returns:
             bs4.BeautifulSoup: BeautifulSoup object representing the requested page's html
         """
 
-        if self._session is None:
-            req = requests.get(url)
-        else:
-            req = self._session.session.get(url)
-        if req.status_code == 429:
-            raise utils.HTTPError("We are being rate-limited. Try again in a while or reduce the number of requests")
-        content = req.content
-        soup = BeautifulSoup(content, "lxml")
+        req = self.get(url)
+        soup = BeautifulSoup(req.content, "lxml")
         return soup
 
     @staticmethod
